/*
   Written 2002 by Josef Soucek <josef.soucek@ct.cz>
   Please read pclink.c for more information.
*/

#include <stdio.h>	// because of printf
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/io.h>
#include "ports.h"

#define true 1
#define false 0

#ifndef DUART
static void sendnibble(unsigned char data);
static unsigned char getnibble(void);
#endif

extern void testtimerandkeyboard();
extern void settimercount(short);

extern unsigned char timeouterr;
short timeout=0;

int commhandle;

#ifdef DUART

/*---------------------------------------------------------------------------*/
/* Init COM port                                                             */
/*---------------------------------------------------------------------------*/
/* 115200 8N1 CTS/RTS */

#define BAUD B115200

int initport(char *commname,int flowctrl)
{
  int flags;
  struct termios options;

  printf("Initializing %s port...\n",commname);
  commhandle = open(commname, O_RDWR | O_NONBLOCK | O_NOCTTY);

  if (commhandle < 0) 
    {
      printf("Cannot open %s\n",commname);
      return false;
    }

  if (tcgetattr(commhandle, &options) < 0)
    {
      printf("Cannot read params %s\n",commname);
      return false;
    }

  /* baud rate */

  cfsetispeed(&options, BAUD);
  cfsetospeed(&options, BAUD);

  /* input options */

  options.c_iflag = 0;

  /* cout options */

  options.c_oflag = 0;

  /* control options */
  /* 8N1 */

  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;

  /* CTS/RTS flow control */

  if (flowctrl)
  {
    printf("Enabling RTS/CTS flow control\n");
    options.c_cflag |= CRTSCTS;
  } else options.c_cflag &= ~CRTSCTS;

  /* line options */
  /* raw input */
 
//  options.c_lflag |= IEXTEN; /* leave enabled to get flow control */
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ISIG);

  /* control character */

//  options.c_cc[VMIN] = 1; /* minimum number of character to read */
  options.c_cc[VMIN] = 0; /* minimum number of character to read */
  options.c_cc[VTIME] = 0; /* time to wait for data */

  if (tcsetattr(commhandle, TCSANOW, &options) < 0)
    {
      printf("Cannot set params %s\n",commname);
      return false;
    }

  /* turn turn O_NONBLOCK again for sure */
  /* O_NONBLOCK = O_NDELAY -> DOESNT CARE WHAT STATE DCD */


  flags = fcntl(commhandle, F_GETFL, 0);
  if (flags < 0) return false;
  if (fcntl(commhandle, F_SETFL, flags & ~O_NONBLOCK | FNDELAY ) < 0) return false;

  /* set the read function to return immediately */

//  fcntl(commhandle, F_SETFL, FNDELAY);

  printf("%s initialized O.K.\n",commname);
  return true;
}

void closeport()
{
  close(commhandle);
}

/*---------------------------------------------------------------------------*/
/* Reads Byte from COM                                                       */
/*---------------------------------------------------------------------------*/
unsigned char getbyte()
{
  unsigned char data;
  ssize_t bytes;

  //fprintf(stderr,"#");
  
  settimercount(timeout);

  while(!timeouterr)
  {
    if ((bytes=read(commhandle,&data,1)) == 1) return data;
//    printf("Data:x%02x; Bytes:x%02x;  ",data,bytes);
    testtimerandkeyboard();
  }
  return false;  /* timeout occur no valid data */
}  

/*---------------------------------------------------------------------------*/
/* Sends Byte to COM                                                         */
/*---------------------------------------------------------------------------*/
void sendbyte(unsigned char data)   /* Send one byte (from 'ByteBuf') to C64 */
{
//  int status;
//  settimercount(1);     /* Fixed timeout */
//  do
//  {
//    testtimerandkeyboard();
//    ioctl(commhandle, TIOCMGET, &status);
//  } while (!(status & TIOCM_CTS) || !timeouterr );
//  } while (!(status & TIOCM_CTS));
//printf("#\n");

//  if (status & TIOCM_CTS) 
    if (write(commhandle,&data,1)!=1)
      printf("Sending error\n");

  timeouterr=0;

  /* there were FIFO overrun in C64 */
  /* CHANNEL A OVERRUN ERROR  SRA bit 4      */

  usleep(1); /* Temporary solution until RTS/CTS handshaking fixed */
}

#else

/*---------------------------------------------------------------------------*/
/* Init parallel port                                                        */
/*---------------------------------------------------------------------------*/

//#define DATA PORT
#define STATUS DATA+1
#define CONTROL DATA+2

int initport(char *commname,int flowctrl)
{
  printf("Initializing parallel port 0x%3x ...\n",DATA);

  if (ioperm(DATA,3,1))
  {
    printf("Access to the port failed...\n");
    printf("The program needs root permisions\n");
    return false;
  }
  outb(inb(DATA) | 0x8,DATA);              /* data3=1 */
  printf("Initialized O.K.\n");
  return true;
}

void closeport(void)
{
  outb(inb(DATA) | 0x8,DATA);              /* data3=1 */
}

unsigned char getnibble(void)
{
  unsigned char data;

  settimercount(timeout);

  while(!timeouterr)
  {
    testtimerandkeyboard();                /* direction bit */
    if (!(inb(STATUS) & 0x80))  break;     /* wait for BUSY=high (0) */
  }
  if (timeouterr) return false;

  outb(inb(DATA) & ~0x8,DATA);             /* data3=0 */
  while(!timeouterr)
  {
    testtimerandkeyboard();
    if (!(inb(STATUS) & 0x40))  break;     /* wait for ACK=0*/
  }
  if (timeouterr) return false;
  data=(inb(STATUS) & 0x38) >> 3;               /* read data2-data0 */
  outb(inb(DATA) | 0x8,DATA);              /* data3=1 */

  /* no settimercount so both wait for ACK in one timeout */

  while(!timeouterr)
  {
    testtimerandkeyboard();
    if (inb(STATUS) & 0x80)                /* test for BUSY=low (1) */
    {
      printf("Error, reversed direction durning GN\n");
      break;
    }
    if (inb(STATUS) & 0x40)  break;        /* wait for ACK=1*/
  }
  return data;
}

void sendnibble(unsigned char data)
{
  settimercount(timeout);

  while(!timeouterr)
  {
    testtimerandkeyboard();                /* direction bit */
    if (inb(STATUS) & 0x80)  break;        /* wait for BUSY=low (1) */
  }
  if (timeouterr) return;

  while(!timeouterr)
  {
    testtimerandkeyboard();
    if (!(inb(STATUS) & 0x40))  break;     /* wait for ACK=0*/
  }
  if (timeouterr) return;
  outb(data & 0x7,DATA);                   /* send data */
  outb(inb(DATA) & ~0x8,DATA);             /* data3=0 */

  while(!timeouterr)
  {
    testtimerandkeyboard();
    if (!(inb(STATUS) & 0x80))             /* test for BUSY=hi (0) */
    {
      printf("Error, reversed direction durning SN\n");
      break;
    }
    if (inb(STATUS) & 0x40)  break;        /* wait for ACK=1*/
  }
  outb(inb(DATA) | 0x8,DATA);              /* data3=1 */
}

unsigned char getbyte()
{
  unsigned char data;
  unsigned char datain;

  data=getnibble();
  if (timeouterr) return false;  /* timeout occur no valid data */
  datain=getnibble();
  if (timeouterr) return false;  /* timeout occur no valid data */
  data+=datain << 3;
  datain=getnibble();
  if (timeouterr) return false;  /* timeout occur no valid data */
  data+=datain << 6;
  return data;
}

void sendbyte(unsigned char data) 
{
  sendnibble(data);
  if (timeouterr) return;
  sendnibble(data >> 3);
  if (timeouterr) return;
  sendnibble(data >> 6);
}

#endif
